/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/tools/gnunet-search-sblock.c 
 * @brief Main function to search for SBlocks on GNUnet.
 * @author Christian Grothoff
 **/

#include "gnunet_afs_esed2.h"

static HashCode160 namespace;
static HashCode160 identifier;

typedef struct {
  HashCode160 * results;
  int resultCount;
} NSSearchClosure;
  
/**
 * Prints the usage information for this command if the user errs.
 * Aborts the program.
 **/
static void printhelp() {
  printf("USAGE: gnunet-search-sblock [OPTIONS] NAMESPACE IDENTIFIER\n"\
	 "-a LEVEL, --anonymity=LEVEL   : "\
	 "set receiver-anonymity enforcing to LEVEL\n"\
	 "-t TIMEOUT                    : "\
	 "wait TIMEOUT seconds before aborting\n"\
	 "-o PREFIX, --output=PREFIX    : "\
	 "write encountered (unencrypted) SBlocks to files\n");
  
  printHelpDefault();
}

/**
 * Parse the options, set the timeout.
 * @param argc the number of options
 * @param argv the option list (including keywords)
 * @return SYSERR if we should exit, OK otherwise
 **/
static int parseOptions(int argc,
			char ** argv) {
  int c;  

  while (1) {
    int option_index = 0;
    static struct GNoption long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "anonymity", 1, 0, 'a' }, 
      { "timeout",   1, 0, 't' },
      { "output",    1, 0, 'o' },
      { 0,0,0,0 }
    };    
    c = GNgetopt_long(argc,
		      argv, 
		      "a:vhdc:L:H:o:t:", 
		      long_options, 
		      &option_index);    
    if (c == -1) 
      break;  /* No more flags to process */
    if (YES == parseDefaultOptions(c, GNoptarg))
      continue;
    switch(c) {
    case 'a': {
      unsigned int receivePolicy;

      if (1 != sscanf(GNoptarg, "%ud", &receivePolicy)) {
        LOG(LOG_FAILURE,
            "FAILURE: You must pass a number to the -a option.\n");
        return -1;
      }
      setConfigurationInt("AFS",
                          "ANONYMITY-RECEIVE",
                          receivePolicy);
      break;
    }
    case 'v': 
      printf("GNUnet v%s, gnunet-search v%s\n",
	     VERSION, 
	     AFS_VERSION);
      return SYSERR;
    case 'h': 
      printhelp(); 
      return SYSERR;
    case 'o':
      FREENONNULL(setConfigurationString("GNUNET-SEARCH-SBLOCK",
      					 "OUTPUT_PREFIX",
					 GNoptarg));
      break;
    case 't': {
      unsigned int timeout;
      if (1 != sscanf(GNoptarg, "%ud", &timeout)) {
	LOG(LOG_FAILURE, 
	    "You must pass a number to the -t option.\n");
	return SYSERR;
      } else {
	setConfigurationInt("AFS",
			    "SEARCHTIMEOUT",
			    timeout);
      }
      break;
    }
    default: 
      LOG(LOG_FAILURE, 
	  "Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      return SYSERR;
    } /* end of parsing commandline */
  } /* while (1) */
  if (argc - GNoptind != 2) {
    fprintf(stderr,
	    "Invalid number of arguments.  "
	    "Call with namespace and identifier (in HEX notation).\n");
    return SYSERR;
  }
  hex2hash((HexName*)argv[GNoptind++],
	   &namespace);
  if (SYSERR == tryhex2hash(argv[GNoptind],
  			    &identifier)) {
     LOG(LOG_DEBUG,
         "DEBUG: namespace ID entered is not in HEX format, using hash of ASCII text (%s).\n",
	  argv[GNoptind]);
     hash(argv[GNoptind], strlen(argv[GNoptind]), &identifier);
  }
  GNoptind++;
  return OK;
}

static void cb(SBlock * sb,
	       NSSearchClosure * sqc) {
  HashCode160 curK;
  int i;
  char * prefix;

  hash(sb, sizeof(SBlock), &curK);
  for (i=0;i<sqc->resultCount;i++)
    if (equalsHashCode160(&curK,
        &sqc->results[i])) {
      LOG(LOG_DEBUG, 
	  "DEBUG: SBlock already seen\n");
      return; /* displayed already */
    }
  GROW(sqc->results,
       sqc->resultCount,
       sqc->resultCount+1);
  memcpy(&sqc->results[sqc->resultCount-1],
         &curK,
         sizeof(HashCode160));
  printSBlock(stdout,
	      sb);

  /* write sblock to file? */
  prefix = getConfigurationString("GNUNET-SEARCH-SBLOCK",
  				  "OUTPUT_PREFIX");
  if(prefix != NULL) {
    char * outfile;

    outfile = MALLOC(strlen(prefix)+16);
    sprintf(outfile, "%s.%03d", prefix, sqc->resultCount-1);
    writeFile(outfile,
    	      sb,
	      sizeof(SBlock),
	      "600");
    FREE(outfile);
    FREE(prefix);
  }
}

static int searchMain(GNUNET_TCP_SOCKET * sock) {
  int ret;
  NSSearchClosure sqc;
	  
  sqc.results = NULL;
  sqc.resultCount = 0;
  ret = searchSBlock(sock,
		     &namespace,
		     &identifier,
		     &testShutdown, 
		     NULL,
		     (NSSearchResultCallback)&cb,
		     &sqc);
  if (ret == SYSERR) 
    printf("Sorry, nothing found.\n");
 
  FREENONNULL(sqc.results);
  return ret;
}

/**
 * The main function to search for SBlocks on GNUnet.
 *
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return 0: ok, -1: error
 **/   
int main(int argc,
	 char ** argv) {
  GNUNET_TCP_SOCKET * sock;
  void * unused;
  pthread_t searchThread;
 
  if (SYSERR == initUtil(argc, argv, &parseOptions))
    return 0;

  sock = getClientSocket();
  initAnonymityPolicy(NULL);
  initializeShutdownHandlers();
  /* order of cron-jobs is important, thus '- cronMILLIS' */
  addCronJob((CronJob)&run_shutdown,
	     cronSECONDS * getConfigurationInt("AFS",
					       "SEARCHTIMEOUT") - cronMILLIS,
	     0, /* no need to repeat */
	     NULL);
  startCron();
  if (0 != PTHREAD_CREATE(&searchThread,
			  (PThreadMain)&searchMain,
			  sock,
			  4 * 1024))
    errexit("FATAL: failed to create search thread.\n");
  wait_for_shutdown();
  closeSocketTemporarily(sock);
  

  stopCron();
  delCronJob((CronJob)&run_shutdown,
	     0,
	     NULL);
  PTHREAD_JOIN(searchThread, &unused);

  doneAnonymityPolicy();
  releaseClientSocket(sock);
  doneUtil();
  doneShutdownHandlers();
  return 0;
}

/* end of gnunet-search-sblock.c */
